open Result
open Monad.MonadList

(* Redundant Rules Method *)
(* 時間がないので、仮のモジュールに定義する *)

(* least fixed point for concat map *)
let rec fix f xs = List.rev (fix0 f xs (Listx.concat_map f xs))
and fix0 f xs = function
  | [] -> xs
  | ys when Listset.subset ys xs -> xs
  | ys -> fix0 f (Listset.union ys xs) (Listx.concat_map f ys)

module I = Index.Make (
  struct
    type t = Rule.t
    let compare = Stdlib.compare
  end
)

let rec chain1 t = function
  | [] -> []
  | lr::rs ->
      let fs = Term.functions t in
      let ds = Rules.defined_symbols [lr] in
      if Listset.subset ds fs then
        I.index lr :: chain1 t rs
      else
        chain1 t rs

(* TRS R -> [(t, R')] where
  - all rules in R' are possible rewritable t
  - t is a term of a rule (l,r) in R
*)
let rec chain rs =
  let chain1L = let* l,r = rs in
                let ss = Listset.diff rs [l,r] in
                [(I.index (l,r), chain1 l ss)] in
  let chain1R = let* l,r = rs in
                let ss = Listset.diff rs [l,r] in
                [(I.index (l,r), chain1 r ss)]  in
  (* search grouped rules (possible rewritable rules) *)
  let f i j = if i = j then [] else
    let css = 
      let* k,cs = chain1R in
      guard (j = k) >>= fun () ->
      return cs in
    Listset.diff (List.hd css) [i] in
  (* merge them  *)
  let llist =
    let* i,cs = chain1L in
    [(fst (I.element i), i), Listx.unique (fix (f i) cs)] in
  let rlist =
    let* i,cs = chain1R in
    [(fst (I.element i), i), Listx.unique (fix (f i) cs)] in
  llist @ rlist

(* remove emulatable rules recursibly *)
let rec remove_rules rs map =
  let indexes =
    let* i,cs = map in
    guard (Listset.subset cs rs) >>= fun () ->
    return i in
  match indexes with
  | [] -> map
  | removable ->
      let xs =
        let* j,ds = map in
        guard (not (List.mem j removable)) >>= fun () ->
        return (j, (Listset.diff ds rs)) in
      remove_rules removable xs

let rec merge_rules map = Listx.unique (merge_rules0 [] map)
and merge_rules0 rs map =
  (* if a rule is necessary, it cannot joinable by other rules *)
  let necessary =
    (let* i,xs = map in
    guard (xs = []) >>= fun () ->
    return i) @ rs in
  let map2 =
    List.filter
      (fun (j,cs) -> not (Listset.subset cs necessary))
      (remove_rules necessary map) in
  match map2 with
  (* no removable rules *)
  | []   -> necessary
  (* there are several removable rules *)
  | need -> 
      match Listset.minimal
        ~leq:(fun (_,cs1) (_,cs2) -> List.length cs1 < List.length cs2)
        need with
      | []       -> rs
      | (j,_)::_ -> let necessary_plus = j :: necessary in
        merge_rules0 necessary_plus (remove_rules necessary_plus need)


let rec remove_redundants b_l2r b_conv k rs =
  begin
    (* compute dependency of lhs and rhs terms in rules *)
    let chainSet = chain rs in
    let minimal = function
      | [] -> []
      | zs -> List.hd @@
        Listset.minimal ~leq:(fun xs ys -> List.length xs < List.length ys) zs in
    (* compute minimal rules (without l->r it self) to l rewrite to r for all l->r *)
    let map2 =
      let* l,r = rs in
      let* (t,i),cs = chainSet in
      guard (i = I.index (l,r) && l = t) >>= fun () ->
      let seqs =
        let* seql,_ = k_join_idx k (List.map I.element cs) [] (l,r) in
        return seql in
      return (i, minimal seqs) in
    let rs1 = if not b_l2r then rs else List.map I.element (merge_rules map2) in
    (* compute minimal rules (without l->r) to join l and r for all l->r *)
    let rs2 = if not b_conv then rs1 else
      List.map I.element (
        merge_rules (
          let* l,r = rs in
          let* (t,i),cs = chainSet in
          guard (i = I.index (l,r) && l = t) >>= fun () ->
          let* (u,j),ds = chainSet in
          guard (j = I.index (l,r) && r = u) >>= fun () ->
            let seqs =
              let csrl = List.map I.element (Listset.diff cs [j]) in
              let dsrl = List.map I.element (Listset.diff ds [i]) in
              let* seql,seqr = k_join_idx k csrl dsrl (l,r) in
              return (Listset.union seql seqr) in
          return (i, minimal seqs)))
    in rs2
  end
and k_join_idx k rs ss e =
  let indexes seqs =
    let* (_,rls,_) = seqs in
    let* rl,_ = rls in
    return (I.index rl) in
  let* seql, seqr = Join.k_join k rs ss e in
  let lindexes = indexes seql in
  let rindexes = indexes seqr in
  return (Listx.unique lindexes, Listx.unique rindexes)


let if_reduced r r' =
  if Listset.equal r r' then
    ()
  else
    begin
      print_endline "NOTE: input TRS is reduced\n";
      Format.printf "original is\n%s@." (Trs.sprint_rules r);
      Format.printf "reduced to\n%s@."  (Trs.sprint_rules r');
    end
